/*
 * Copyright (c) 1997, 1998, 1999
 *      Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *      $Id: wan_ng_iface.c,v 1.15 2007/12/10 17:47:47 sangoma Exp $
 */


/*
 ****************************************************************************
 * sdla_ng.c   WANPIPE(tm) Multiprotocol WAN Link Driver.  Main module.
 *
 * Author:      Alex Feldman    <al.feldman@sangoma.com>
 *
 * ===========================================================================
 * Feb 09, 2007 Joel M Pareja	Netgraph failover support - ng_one2many
 * Sep 10, 2003 Alex Feldman    Initial version.
 ****************************************************************************
*/
#include <wanpipe_version.h>
#include <wanpipe_includes.h>
#include <wanpipe.h>        /* WANPIPE common user API definitions */
#include <if_wanpipe_common.h>
#include <wanpipe_iface.h>

#define NG_WAN_NODE_TYPE        "wp"
#define NG_WAN_COOKIE		917786112
#define NG_WAN_HOOK_DEBUG       "debug"
#define NG_WAN_HOOK_UPPER	"upper"
#define NG_WAN_HOOK_LOWER	"lower"

/* Netgraph control messages */
enum {
        NGM_WAN_GET_PROMISC = 1,      /* get node's promiscuous mode bit */
        NGM_WAN_SET_PROMISC,          /* enable/disable promiscuous mode */
        NGM_WAN_GET_AUTOSRC,          /* get source address override */
        NGM_WAN_SET_AUTOSRC,          /* enable/disable src addr override */
};


/* Function interface between LIP layer and kernel */
extern wan_iface_t wan_iface;

static ng_constructor_t wan_ng_constructor;
static ng_rcvmsg_t      wan_ng_rcvmsg;
static ng_shutdown_t    wan_ng_shutdown;
static ng_newhook_t     wan_ng_newhook;
static ng_connect_t     wan_ng_connect;
static ng_rcvdata_t     wan_ng_rcvdata;
static ng_disconnect_t  wan_ng_disconnect;

/* List of commands and how to convert arguments to/from ASCII */
static const struct ng_cmdlist wan_ng_cmdlist[] = {
        {
          NG_WAN_COOKIE,
          NGM_WAN_GET_PROMISC,
          "getpromisc",
          NULL,
          &ng_parse_int32_type
        },
        {
          NG_WAN_COOKIE,
          NGM_WAN_SET_PROMISC,
          "setpromisc",
          &ng_parse_int32_type,
          NULL
        },
        {
          NG_WAN_COOKIE,
          NGM_WAN_GET_AUTOSRC,
          "getautosrc",
          NULL,
          &ng_parse_int32_type
        },
        {
          NG_WAN_COOKIE,
          NGM_WAN_SET_AUTOSRC,
          "setautosrc",
          &ng_parse_int32_type,
          NULL
        },
        { 0 }
};


static struct ng_type typestruct = {
        .version =         NG_ABI_VERSION,
        .name =            NG_WAN_NODE_TYPE,
        .constructor =     wan_ng_constructor,
        .rcvmsg =          wan_ng_rcvmsg,
        .shutdown =        wan_ng_shutdown,
        .newhook =         wan_ng_newhook,
        .connect =         wan_ng_connect,
        .rcvdata =         wan_ng_rcvdata,
        .disconnect =      wan_ng_disconnect,
	.cmdlist =         wan_ng_cmdlist,
};
MODULE_DEPEND(wanpipe, netgraph, NG_ABI_VERSION,NG_ABI_VERSION,NG_ABI_VERSION);
/*NETGRAPH_INIT(wp, &typestruct);*/

int		wan_ng_init(void);
int		wan_ng_destroy(void);
int		wan_ng_create(void *arg);
int		wan_ng_remove(void *arg);
static int	wan_ng_rcv_lower(node_p node, struct mbuf *m);
static int	wan_ng_rcv_upper(node_p node, struct mbuf *m);
static int	wan_ng_up(wanpipe_common_t *common);
static int	wan_ng_down(wanpipe_common_t *common);
void		wan_ng_link_state(wanpipe_common_t *common, int state);
int		wan_ng_input(wanpipe_common_t *common, netskb_t *);

int wan_ng_create(void *arg)
{
	wanpipe_common_t	*common = (wanpipe_common_t*)arg;
	int			err;

	DEBUG_TEST("%s: Creating new NETGRAPH node...\n",
				wan_netif_name(common->dev));
	if ((err = ng_make_node_common(&typestruct, &common->ng_node)) != 0){
		DEBUG_EVENT("%s: Failed to create NETGRAPH node (err=%d)!\n",
                                  wan_netif_name(common->dev), err);
		return -EINVAL;
	}
	NG_NODE_SET_PRIVATE(common->ng_node, common);
	strcpy(common->ng_nodename, wan_netif_name(common->dev));
	if ((err = ng_name_node(common->ng_node, wan_netif_name(common->dev)))) {
		DEBUG_EVENT("%s: Failed to assign a netgraph node a name (err = %d)!\n",
                           wan_netif_name(common->dev), err);
		ng_rmnode_self(common->ng_node);
		NG_NODE_UNREF(common->ng_node);
		common->ng_node = NULL;
		return -EINVAL;
	}

	common->lo_queue.ifq_maxlen = IFQ_MAXLEN;
	common->hi_queue.ifq_maxlen = IFQ_MAXLEN;
	mtx_init(&common->lo_queue.ifq_mtx, "wp_queue_lo", NULL, MTX_DEF);
	mtx_init(&common->hi_queue.ifq_mtx, "wp_queue_hi", NULL, MTX_DEF);
	return 0;
}

int wan_ng_remove(void *arg)
{
	wanpipe_common_t	*common = (wanpipe_common_t*)arg;

	DEBUG_EVENT("%s: Removing NETGRAPH node...\n",
				wan_netif_name(common->dev));
	if (common->ng_node){
		ng_rmnode_self(common->ng_node);
		mtx_destroy (&common->lo_queue.ifq_mtx);
		mtx_destroy (&common->hi_queue.ifq_mtx);
		NG_NODE_UNREF(common->ng_node);
		common->ng_node = NULL;
	}
	return 0;
}

/***********************************************************************
 * This section contains the methods for the Netgraph interface
 ***********************************************************************/
/*
 * It is not possible or allowable to create a node of this type.
 * If the hardware exists, it will already have created it. */
static int wan_ng_constructor(node_p node)
{
	WAN_DEBUG_FUNC_LINE;
        return (EINVAL);
}

/*
 * give our ok for a hook to be added...
 * If oe are not running this should kick the device into life.
 * The hook's private info points to our stash of info about that
 * channel. */
static int wan_ng_newhook(node_p node, hook_p hook, const char *name)
{
        wanpipe_common_t	*common = NG_NODE_PRIVATE(node);
	int			s;

	DEBUG_NG("%s: Connect new hook %s!\n",
					wan_netif_name(common->dev), name);
        /*
         * check if it's our friend the debug hook
         */
        if (strcmp(name, NG_WAN_HOOK_DEBUG) == 0) {
                NG_HOOK_SET_PRIVATE(hook, NULL);
                common->ng_debug_hook = hook;
                return (0);
        }

	if (!strcmp(name, NG_WAN_HOOK_UPPER)) {
		common->ng_upper_hook = hook;
		common->ng_upperhooks++;
	} else if (!strcmp(name, NG_WAN_HOOK_LOWER)) {
		common->ng_lower_hook = hook;
		common->ng_lowerhooks++;
	} else {
		DEBUG_EVENT("%s: invalid hook %s!\n",
					wan_netif_name(common->dev), name);
		return (EINVAL);
	}
        NG_HOOK_SET_PRIVATE(hook, common);
	s = splhigh();
        wan_ng_up(common);
	splx(s);
        return (0);
}



/*
 * incoming messages.
 * Just respond to the generic TEXT_STATUS message
 */
static  int
wan_ng_rcvmsg(node_p node, item_p item, hook_p lasthook)
{
        wanpipe_common_t	*common = NG_NODE_PRIVATE(node);
        struct ng_mesg		*resp_msg = NULL;
        struct ng_mesg		*msg;
        u_char			want;
        int			error = 0;

        NGI_GET_MSG(item, msg);
        switch (msg->header.typecookie) {
        case NGM_GENERIC_COOKIE:
                switch(msg->header.cmd) {
                case NGM_TEXT_STATUS:
                        {
                        char    *arg;
                        int pos = 0;
                        int resplen;

                        resplen = sizeof(struct ng_mesg) + NG_TEXTRESPONSE;
                        NG_MKRESPONSE(resp_msg, msg, resplen, M_NOWAIT);
                        if (resp_msg == NULL) {
                                error = ENOMEM;
                                break;
                        }
			bzero (resp_msg, resplen);
                        arg = (resp_msg)->data;

                        /*
                         * Put in the throughput information.
                         */
                        pos = sprintf(arg, "%ld bytes in, %ld bytes out\n",
                                        common->dev->if_ibytes,
					common->dev->if_obytes);
                        pos += sprintf(arg + pos,
                                "%ld output errors\n",
                                        common->dev->if_oerrors);
                        pos += sprintf(arg + pos,
                                "ierrors = %ld", common->dev->if_ierrors);
                        resp_msg->header.arglen = pos+1;
                        }
                        break;
               default:
                        error = EINVAL;
                        break;
                }
                break;
	case NG_WAN_COOKIE:
		DEBUG_TEST("%s: Received WAN cookie type!\n",
					wan_netif_name(common->dev));
		switch(msg->header.cmd) {
		case NGM_WAN_GET_PROMISC:
			NG_MKRESPONSE(resp_msg, msg, sizeof(u_int32_t), M_NOWAIT);
			if (resp_msg == NULL) {
				error = ENOMEM;
				break;
			}
			*((u_int32_t *)resp_msg->data) = common->ng_promisc;
			break;
		case NGM_WAN_SET_PROMISC:
                        if (msg->header.arglen != sizeof(u_int32_t)) {
                                error = EINVAL;
                                break;
                        }
                        want = !!*((u_int32_t *)msg->data);
                        if (want ^ common->ng_promisc) {
				if ((error = ifpromisc(common->dev, want)) != 0){
					common->ng_promisc = want;
				}
			}
			break;
		case NGM_WAN_GET_AUTOSRC:
			NG_MKRESPONSE(resp_msg, msg, sizeof(u_int32_t), M_NOWAIT);
			if (resp_msg == NULL) {
                                error = ENOMEM;
                                break;
                        }
                        *((u_int32_t *)resp_msg->data) = common->ng_autoSrcAddr;
			break;
		case NGM_WAN_SET_AUTOSRC:
			if (msg->header.arglen != sizeof(u_int32_t)) {
                                error = EINVAL;
                                break;
                        }
                        common->ng_autoSrcAddr = !!*((u_int32_t *)msg->data);
                        break;
		default:
			error = EINVAL;
			break;
		};
		break;
        default:
		DEBUG_TEST("%s:%d: cmd=%X\n",__FUNCTION__,__LINE__,
        			msg->header.typecookie);
                //error = EINVAL;
                break;
        }
        NG_RESPOND_MSG(error, node, item, resp_msg);
        NG_FREE_MSG(msg);
        return (error);
}

/*
 * get data from another node and transmit it to the correct channel
 */
static int wan_ng_rcvdata(hook_p hook, item_p item)
{
        node_p			node = NG_HOOK_NODE(hook);
        struct mbuf		*m;
        wanpipe_common_t	*common = NG_NODE_PRIVATE(node);
        int			error = 0;

        NGI_GET_M(item, m);
        NGI_GET_META(item, meta);
        NG_FREE_ITEM(item);

        /*
         * data doesn't come in from just anywhere (e.g control hook)
         */
        if (NG_HOOK_PRIVATE(hook) == NULL || common == NULL) {
                error = ENETDOWN;
                goto bad;
        }

 	if (hook == common->ng_lower_hook) {
		return (wan_ng_rcv_lower(node, m));
	}
	if (hook == common->ng_upper_hook) {
		return (wan_ng_rcv_upper(node, m));
	}

	/* invalid hook */
	DEBUG_NG("%s: ERROR: received packet NOT belonging to an hook!\n",
					wan_netif_name(common->dev));
	return (0);

bad:
        /*
         * It was an error case.
         * check if we need to free the mbuf, and then return the error
         */
        NG_FREE_M(m);
        NG_FREE_META(meta);
	return (error);
}

/*
 * Process data set for tranmission to device.
 */
static int  wan_ng_rcv_lower(node_p node, struct mbuf *m)
{
        wanpipe_common_t	*common = NG_NODE_PRIVATE(node);

        WAN_ASSERT(common == NULL);
	/*
         * Now queue the data for when it can be sent
         */
	if (common->ng_running){
		if (m){
			struct sockaddr	dst;
			int		err;
			dst.sa_family = AF_INET;
			if (wan_skb2buffer((void**)&m)){
				DEBUG_EVENT(
				"%s: Failed to correct mbuf list!\n",
					wan_netif_name(common->dev));
				IF_DROP(&common->dev->if_snd);	//IF_DROP(q);
				wan_skb_free(m);
				return ENOBUFS;
			}
			DEBUG_TX("%s: Sending down %d bytes\n",
					wan_netif_name(common->dev),
					wan_skb_len(m));
			wan_bpf_report(common->dev, m, 1, WAN_BPF_DIR_OUT);
			err = common->iface.output(common->dev, m, &dst, NULL);
			if (err == 0){
			}else{
				DEBUG_TX(
				"%s: Failed to send down packet!\n",
						wan_netif_name(common->dev));
				IF_DROP(&common->dev->if_snd);
				wan_skb_free(m);
				return ENETDOWN;
			}
		}
	}
        return (0);
}

static int  wan_ng_rcv_upper(node_p node, struct mbuf *m)
{
        wanpipe_common_t	*common = NG_NODE_PRIVATE(node);
        netdevice_t*            ifp = NULL;
	int                     error = 0;

        if ((ifp = common->dev) == NULL){
        	error = EINVAL;
                goto bad;
	}
	/* If we are sending it up, remove our M_FLAG */
	wan_skb_clear_mark(m);
	//wan_bpf_report(common->dev, m, 1, WAN_BPF_DIR_IN);
	if (common->is_netdev){
		netisr_queue(
			(m->m_flags & M_PROTO1)?NETISR_IPX:NETISR_IP, m); 
	}else{
		WAN_SPPP_INPUT(ifp, m);
	}
	return (0);

bad:
        /*
         * It was an error case.
         * check if we need to free the mbuf, and then return the error
         */
        NG_FREE_M(m);
        return (error);
}

static int wan_ng_up(wanpipe_common_t *common)
{
	common->ng_running = 1;
	return 0;
}
static int wan_ng_down(wanpipe_common_t *common)
{
	common->ng_running = 0;
	return 0;
}

/*
 * do local shutdown processing..
 * this node will refuse to go away, unless the hardware says to..
 * don't unref the node, or remove our name. just clear our links up.
 */
static int wan_ng_shutdown(node_p node)
{
        wanpipe_common_t	*common = NG_NODE_PRIVATE(node);

	DEBUG_NG("%s: Shutdown the hook!\n",
					wan_netif_name(common->dev));

	wan_ng_down(common);
        return (0);
}

/* already linked */
static int wan_ng_connect(hook_p hook)
{
        /* be really amiable and just say "YUP that's OK by me! " */
#if defined(WAN_DEBUG_NG)
	wanpipe_common_t	*common = NG_NODE_PRIVATE (NG_HOOK_NODE (hook));
#endif

	DEBUG_NG("%s: Connect the hook!\n",
					wan_netif_name(common->dev));

	return 0;
}

/*
 * Notify on hook disconnection (destruction)
 *
 * Invalidate the private data associated with this dlci.
 * For this type, removal of the last link resets tries to destroy the node.
 * As the device still exists, the shutdown method will not actually
 * destroy the node, but reset the device and leave it 'fresh' :)
 *
 * The node removal code will remove all references except that owned by the
 * driver.
 */
static int wan_ng_disconnect(hook_p hook)
{
        wanpipe_common_t	*common = NG_NODE_PRIVATE( NG_HOOK_NODE(hook) );
        int			s;
        /*
         * If it's the data hook, then free resources etc.
         */

	WAN_ASSERT(common == NULL);
	DEBUG_NG("%s: Disconnect the hook!\n",
					wan_netif_name(common->dev));

	if (hook == common->ng_upper_hook) {
		s = splimp();
		common->ng_upperhooks--;
		common->ng_upper_hook = NULL;
		splx(s);
	} else if (hook == common->ng_lower_hook) {
		s = splimp();
		common->ng_lowerhooks--;
		common->ng_lower_hook = NULL;
		splx(s);
	} else {
		/* what? */
	}
 	if ((common->ng_lowerhooks == 0) && (common->ng_upperhooks == 0)){
		wan_ng_down(common);
	}
	return (0);
}

/*
 * Notify ng_one2many node about link event.
 */
void
wan_ng_link_state(wanpipe_common_t *common, int state)
{
        struct ng_mesg *msg;
        int cmd, dummy_error = 0;

        if (common->ng_lower_hook == NULL)
                return;

        if (state == WAN_CONNECTED)
                cmd = NGM_LINK_IS_UP;
        else if (state == WAN_DISCONNECTED)
                cmd = NGM_LINK_IS_DOWN;
        else
                return;

        NG_MKMESSAGE(msg, NGM_FLOW_COOKIE, cmd, 0, M_NOWAIT);
        if (msg != NULL)
        {
                NG_SEND_MSG_HOOK(dummy_error, common->ng_node, msg,
                                 common->ng_lower_hook, 0);
        }

	return;
}

/*
 * called during bootup
 * or LKM loading to put this type into the list of known modules
 */
int wan_ng_init(void)
{
	int err = 0;
	
	if ((err = ng_newtype(&typestruct))){
		DEBUG_EVENT("ERROR: Failed to register ng_wp (err=h type (%d)!\n", err);
		return -EINVAL;
	}
	return 0;
}

int wan_ng_destroy(void)
{
	ng_rmtype(&typestruct);
	return 0;
}

int wan_ng_input(wanpipe_common_t *common, netskb_t *skb)
{
	int error;

	if (common->ng_lower_hook == NULL){
		return -EINVAL;
	}
	DEBUG_RX("%s: Sending to lower hook packet %d bytes (%p)!\n", 
					wan_netif_name(common->dev),
					wan_skb_len(skb));
	NG_SEND_DATA_ONLY(error, 
			  common->ng_lower_hook, 
			  skb);
	return 0;
}
